// ------------------------------------------------------------------
// test-changement-governor - programme de test du changement
//                            dynamique de governor.
//
// (c) 2011 Christophe BLAESS - Tous droits reserves.
//
// Exemple pour un post du blog http://www.blaess.fr/christophe/
// ------------------------------------------------------------------

#define _GNU_SOURCE // Pour avoir pthread_attr_setaffinity_np()

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>

// Parametrage du programme :
//
//   Les CPUs (ou coeurs) sur lesquels les threads vont s'executer
#    define CPU_THREAD_1             0
#    define CPU_THREAD_2             1
//   Les priorites temps-reel des threads (0 = pas temps-reel)
#    define PRIORITE_THREAD_1        10
#    define PRIORITE_THREAD_2        10
//   Le nombre d'iteration dans une boucle active
#    define NB_ITERATIONS_PAR_BOUCLE 200000
//   Le nombre de boucles actives a executer
#    define NB_BOUCLES               100

static void * thread_1 (void * arg);
static void * thread_2 (void * arg);
static pthread_barrier_t barriere;

static char * governor_from = NULL;
static char * governor_to   = NULL;

int main(int argc, char * argv[])
{
	cpu_set_t  cpuset;
	pthread_t thr_1;
	pthread_t thr_2;
	pthread_attr_t attr;

	struct sched_param param;

	if (argc != 3) {
		fprintf(stderr, "usage: %s governor_1 governor_2\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	governor_from = argv[1];
	governor_to   = argv[2];
	
	pthread_barrier_init(& barriere, NULL, 2);

	/* Thread modifiant la frequence CPU */
	pthread_attr_init(& attr);
	
	CPU_ZERO(& cpuset);
	CPU_SET(CPU_THREAD_1, & cpuset);
	if (pthread_attr_setaffinity_np(& attr, sizeof(cpuset), & cpuset) != 0) {
		perror("pthread_attr_setaffinity_np");
		exit(2);
	}
	if (PRIORITE_THREAD_1 != 0) {
		if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO) != 0) {	
			perror("pthread_attr_setschedpolicy");
			exit(2);
		}
		param.sched_priority = PRIORITE_THREAD_1;
		if (pthread_attr_setschedparam(&attr, &param) != 0) {
			perror("pthread_attr_setschedparam");
			exit(2);
		}
		if (pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED) != 0) {
			perror("pthread_attr_setinheritsched");
			exit(2);
		}
	}
	if ((errno = pthread_create(& thr_1, &attr, thread_1, NULL)) != 0) {
		perror("pthread_create");
		exit(2);
	}
	/* Thread effectuant l'echantillonnage */
	pthread_attr_init(& attr);
	CPU_ZERO(& cpuset);
	CPU_SET(CPU_THREAD_2, & cpuset);
	if (pthread_attr_setaffinity_np(& attr, sizeof(cpuset), & cpuset) != 0) {
		perror("pthread_attr_setaffinity_np");
		exit(2);
	}
	if (PRIORITE_THREAD_2 != 0) {
		if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO) != 0) {
			perror("pthread_attr_setschedpolicy");
			exit(2);
		}
		param.sched_priority = PRIORITE_THREAD_2;
		if (pthread_attr_setschedparam(&attr, &param) != 0) {
			perror("pthread_attr_setschedparam");
			exit(2);
		}
		// sans tenir compte de l'ordonnancement du thread main()
		if (pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED) != 0) {
			perror("pthread_attr_setinheritsched");
			exit(2);
		}
	}
	if ((errno = pthread_create(& thr_2, &attr, thread_2, NULL)) != 0) {
		perror("pthread_create");
		exit(2);
	}
	pthread_join(thr_1, NULL);
	pthread_join(thr_2, NULL);
	return EXIT_SUCCESS;
}

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;



static void * thread_1 (void * unused)
{
	// Thread modifiant la frequence du CPU_THREAD_2
	int fd;
	char buffer[256];
	snprintf(buffer, 256, "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_governor", CPU_THREAD_2);

	fd = open(buffer, O_RDWR);
	if (fd < 0) {
		perror(buffer);
		exit(2);
	}
	
	snprintf(buffer, 256, "%s\n", governor_from);
	if (write(fd, buffer, strlen(buffer)) < 0) {
		perror(buffer);
		exit(2);
	}
	// stabilisation du governor
	sleep(1);
	
	snprintf(buffer, 256, "%s\n", governor_to);

	// Synchronisation avec demarrage thread 2
	pthread_barrier_wait(& barriere);

	// Attente mi-parcours travail thread 2
	pthread_mutex_lock(& mutex);
	if (write(fd, buffer, strlen(buffer)) < 0) {
		perror(buffer);
		exit(2);
	}
	close(fd);
	pthread_exit(NULL);
}


static void * thread_2(void * unused)
{
	// Thread echantillonnant des boucles
	struct timeval tv[NB_BOUCLES+1];
	int i;
	int j;
	long long int duree;

	// Verrouillage initial du mutex de mi-parcours
	pthread_mutex_lock(& mutex);

	// Attente de demarrage synchronise avec thread 1
	pthread_barrier_wait(& barriere);

	for (i = 0; i < NB_BOUCLES / 2; i ++) {
		gettimeofday(& (tv[i]), NULL);
		for (j = 0; j < NB_ITERATIONS_PAR_BOUCLE; j ++)
			;
	}
	// Mi-parcours
	pthread_mutex_unlock(& mutex);
	for (; i < NB_BOUCLES; i ++) {
		gettimeofday(& (tv[i]), NULL);
		for (j = 0; j < NB_ITERATIONS_PAR_BOUCLE; j ++)
			;
	}
	gettimeofday(& (tv[i]), NULL);
	
	for (i = 0; i < NB_BOUCLES; i ++) {
		duree  = tv[i+1].tv_sec - tv[i].tv_sec;
		duree *= 1000000;
		duree += tv[i+1].tv_usec - tv[i].tv_usec;
		fprintf(stdout, "%4d %6lld\n", i, duree);
	}
	pthread_exit(0);
}

